#include <QApplication>
#if QT_VERSION >= 0x050000 && defined(Q_OS_WIN)
#  define WINDOWS_NATIVE_EVENT_FILTER
#  include <QAbstractNativeEventFilter>
#  include <qt_windows.h>
#endif
#include <QMainWindow>
#include <QMenuBar>
#include <QMenu>
#include <QAction>
#include <QVector>
#include <QPainter>
#include <QColor>
#include <QDebug>
#include <QStringList>
#include <QLibraryInfo>
#include <QDesktopWidget>
#include <QMouseEvent>
#include <QShortcut>

#ifdef WINDOWS_NATIVE_EVENT_FILTER

class NativeEventFilter : public QAbstractNativeEventFilter {
public:
    bool nativeEventFilter(const QByteArray &eventType, void *message, long *result);
};

bool NativeEventFilter::nativeEventFilter(const QByteArray & /* eventType */, void *message, long * /* result */)
{
    enum : quint64 { signatureMask = 0xffffff00, miWpSignature = 0xff515700 };

    static int n = 0;
    const MSG *msg = static_cast<const MSG *>(message);
    if (msg->message >= WM_MOUSEFIRST && msg->message <= WM_MOUSELEAVE && msg->message != WM_MOUSEMOVE) {
        const quint64 extraInfo = GetMessageExtraInfo();
        QDebug d = qDebug().nospace();
        d << '#' << n++ << " Native mouse " << hex <<showbase << msg->message << " extra=" << extraInfo;
        if (extraInfo) {
            if ((extraInfo & signatureMask) == miWpSignature)
                d << " [signature]";
            if (extraInfo & 0x80)
                d << " [touch]";
            d << " cursorId=" << (extraInfo & 0x7F);
        }
    }
    return false;
}

#endif // Qt 5 Win

static inline void dumpObject(QDebug &d, const QObject *o)
{
    d << o->metaObject()->className();
    const QString on = o->objectName();
    if (!on.isEmpty())
        d << '/' << on;
}

static void dumpMouseEvent(const char *which, const QObject *o, const QMouseEvent *ev)
{
    static int n = 0;
    QDebug d = qDebug().nospace();
    d << '#' << n++ << ' ' << which;
#if QT_VERSION >= 0x050300
    if (const int fls = ev->flags())
        d << " flags=" << fls;
    if (const int src = ev->source())
        d << " src=" << src;
#endif
    d << " pos=" << ev->pos().x() << ',' << ev->pos().y()
      << " global="
      << ' ' << ev->globalPos().x() << ',' << ev->globalPos().y()
      << ' ';
    dumpObject(d, o);
}

static void dumpTouchEvent(const char *which, const QObject *o, const QTouchEvent *ev)
{
    static int n = 0;
    const QList<QTouchEvent::TouchPoint> pts = ev->touchPoints();

    QDebug d = qDebug().nospace();
    if (!n) {
#if QT_VERSION >= 0x050000
        const QTouchDevice *dev = ev->device();
        d << "QTouchDevice: " << dev->name() << " type=" << dev->type()
          << " max pts: " << dev->maximumTouchPoints()
          << " caps: " << showbase << hex << int(dev->capabilities()) << noshowbase << dec << '\n';
#endif
    }

    d << '#' << n++ << ' ' << which << pts.size() << " points, states=" << ev->touchPointStates();
#if QT_VERSION >= 0x050000
    if (!pts.isEmpty()) {
        if (const int p = pts.front().flags())
            d << "pt flags=" << p;
    }
#endif
    d << ' ';
    dumpObject(d, o);
}

static void dumpTabletEvent(const char *which, const QObject *o, const QTabletEvent *ev)
{
    static int n = 0;
    QDebug d = qDebug().nospace();
    d << '#' << n++ << ' ' << which << " dev=" << ev->device()
      << " ptr=" << ev->pointerType() << " pos="
#if QT_VERSION >= 0x050000
      << ev->posF().x() << ',' << ev->posF().y()
#else
      << ev->pos().x() << ',' << ev->pos().y()
#endif
     << " global="
#if QT_VERSION >= 0x050000
          << ev->globalPosF().x() << ',' << ev->globalPosF().y()
#else
          << ev->globalPos().x() << ',' << ev->globalPos().y()
#endif
             ;
    if (const qreal pres = ev->pressure())
        d << " press=" << pres;
    if (const int z = ev->z())
        d << " z=" << z;
    if (const qreal tp = ev->tangentialPressure())
        d << " tangentialPressure=" << tp;
    if (const int r = ev->rotation())
        d << " rot=" << r;
    if (ev->xTilt() || ev->yTilt())
        d << " tilt " << ev->xTilt() << ',' << ev->yTilt();
#if QT_VERSION >= 0x050000
    d << " time=" << ev->timestamp();
#endif
    d << ' ';
    dumpObject(d, o);
}

class EventDumper: public QObject {
public:
    EventDumper(QObject *p=0): QObject( p ) {}

    bool eventFilter(QObject *o, QEvent *ev) {
        const QEvent::Type et = ev->type();
        switch (et) {
        case QEvent::MouseButtonDblClick:
            dumpMouseEvent("QMouseButtonDblClick ", o, static_cast<const QMouseEvent *>(ev));
            break;
        case QEvent::MouseButtonPress:
            dumpMouseEvent("QMouseButtonPress    ", o, static_cast<const QMouseEvent *>(ev));
            break;
        case QEvent::MouseButtonRelease:
            dumpMouseEvent("QMouseButtonRelease  ", o, static_cast<const QMouseEvent *>(ev));
            break;
        case QEvent::TouchBegin:
            dumpTouchEvent("QTouchBegin          ", o, static_cast<const QTouchEvent *>(ev));
            break;
        case QEvent::TouchUpdate:
            dumpTouchEvent("QTouchUpdate         ", o, static_cast<const QTouchEvent *>(ev));
            break;
        case QEvent::TouchEnd:
            dumpTouchEvent("QTouchEnd            ", o, static_cast<const QTouchEvent *>(ev));
            break;
        case QEvent::TabletEnterProximity:
            dumpTabletEvent("QTabletEnterProximity", o, static_cast<const QTabletEvent *>(ev));
            break;
        case QEvent::TabletLeaveProximity:
            dumpTabletEvent("QTabletLeaveProximity", o, static_cast<const QTabletEvent *>(ev));
            break;
        case QEvent::TabletPress:
            dumpTabletEvent("QTabletPress         ", o, static_cast<const QTabletEvent *>(ev));
            break;
        case QEvent::TabletRelease:
            dumpTabletEvent("QTabletRelease       ", o, static_cast<const QTabletEvent *>(ev));
            break;
        default:
            break;
        }
        return QObject::eventFilter(o, ev );
    }
};

enum TabletPointType {
    TabletButtonPress,
    TabletButtonRelease,
    TabletMove,
    TabletDraw
};

struct TabletPoint
{
    TabletPoint(const QPoint &p = QPoint(), TabletPointType t = TabletMove) : pos(p), type(t) {}

    QPoint pos;
    TabletPointType type;
};

class PaintingWidget : public QWidget
{
    Q_OBJECT
public:
    PaintingWidget(QWidget *p = 0) : QWidget(p),m_lastIsTabletMove(false) {}

public slots:
    void clearPoints() { m_points.clear(); update(); }

protected:
    void tabletEvent(QTabletEvent *);
    void paintEvent(QPaintEvent *);

private:
    QVector<TabletPoint> m_points;
    bool m_lastIsTabletMove;
};

void PaintingWidget::paintEvent(QPaintEvent *)
{
    QPainter p(this);
    const QRect geom = QRect(QPoint(0, 0), size());
    p.fillRect(geom, Qt::white);
    p.drawRect(QRect(geom.topLeft(), geom.bottomRight() - QPoint(1,1)));
    foreach (const TabletPoint &t, m_points) {
        if (geom.contains(t.pos)) {
              QPainterPath pp;
              pp.addEllipse(t.pos, 5, 5);
              switch (t.type) {
              case TabletButtonPress:
                  p.fillPath(pp, Qt::black);
                  break;
              case TabletButtonRelease:
                  p.fillPath(pp, Qt::red);
                  break;
              case TabletMove:
                  p.drawPath(pp);
                  break;
              case TabletDraw:
                  p.fillPath(pp, Qt::blue);
                  break;
              }
          }
    }
}

void PaintingWidget::tabletEvent(QTabletEvent *event)
{
    const bool lastWasTabletMove = m_lastIsTabletMove;
    m_lastIsTabletMove =  event->type() == QEvent::TabletMove;
    QWidget::tabletEvent(event);
    switch (event->type()) {
    case QEvent::TabletMove:
        if (!lastWasTabletMove) {
            m_points.push_back(TabletPoint(event->pos(), event->pressure() ? TabletDraw : TabletMove));
            update();
        }
        break;
    case QEvent::TabletPress:
        m_points.push_back(TabletPoint(event->pos(), TabletButtonPress));
        update();
        break;
    case QEvent::TabletRelease:
        m_points.push_back(TabletPoint(event->pos(), TabletButtonRelease));
        update();
        break;
    default:
        break;
    }
}

class MainWindow : public QMainWindow {
public:
    MainWindow();
};

MainWindow::MainWindow()
{
    setWindowTitle(QT_VERSION_STR);
    setObjectName(QLatin1String("mainWindow"));

    PaintingWidget *pw = new PaintingWidget(this);
    pw->setObjectName("paintingWidget");
    setCentralWidget(pw);

    QMenu *fileMenu = menuBar()->addMenu("File");
    QObject::connect(fileMenu->addAction("Clear"), SIGNAL(triggered()),
                     pw, SLOT(clearPoints()));
    QAction *quitAction = fileMenu->addAction("Quit");
    QObject::connect(quitAction, SIGNAL(triggered()), qApp, SLOT(quit()));
    quitAction->setShortcut(Qt::CTRL + Qt::Key_Q);
}

int main(int argc, char *argv[])
{
    QApplication app(argc, argv);
    QStringList args;
    for (int a = 1; a < argc ; ++a)
        args.append(argv[a]);

    if (!args.contains("-e"))
        app.installEventFilter(new EventDumper(&app));
#if QT_VERSION >= 0x050000
    if (args.contains("-nf"))
        app.installNativeEventFilter(new NativeEventFilter);
#endif

    qDebug()
#if QT_VERSION >= 0x050300
             << QLibraryInfo::build()
#else
             << QT_VERSION_STR
#endif
#if QT_VERSION >= 0x050000
             << QGuiApplication::platformName()
#endif
             << args;

    MainWindow w;
    if (args.contains("-tw")) {
        w.centralWidget()->setAttribute(Qt::WA_AcceptTouchEvents);
        qDebug() << "Qt::WA_AcceptTouchEvents";
    }
    if (!args.contains("-t"))
        w.setWindowFlags(w.windowFlags() | Qt::WindowStaysOnTopHint);

    const QRect ag = app.desktop()->availableGeometry();
    w.resize(ag.width() / 4, ag.height() /4);
    w.setWindowTitle(QT_VERSION_STR);
    w.move(ag.left() + ag.width() / 4,
           ag.top() + ag.height() / 4);
    w.show();
    return app.exec();
}

#include "main.moc"
